require 5.004;
use strict;

my $NAME    = 'PCXFireWall';
my $DISTNAME= "libpcxfirewall-rules-perl";
my $pcx_firewallVersion = "2.17";
my $fileTempVersion = "0.12";
my $xmlSaxVersion = "0.10";
my $xmlLibXMLVersion = "1.40";

# Check for non-standard modules that are used by this library.
$| = 1;
my $missing_modules = 0;

print "Checking for PCXFireWall.........";
eval {
    require PCXFireWall::Config;
    PCXFireWall::Config->VERSION($pcx_firewallVersion);
};
if ($@) {
    print " failed\n";
    $missing_modules++;
    print <<EOT;
$@
The PCXFireWall module version $pcx_firewallVersion must be installed.

EOT
    sleep(2);  # Don't hurry too much
} else {
    print " ok\n";
}

print "Checking for File::Temp.........";
eval {
    require File::Temp;
    File::Temp->VERSION($fileTempVersion);
};
if ($@) {
    print " failed\n";
    $missing_modules++;
    print <<EOT;
$@
The File::Temp module version $fileTempVersion must be installed.

EOT
    sleep(2);  # Don't hurry too much
} else {
    print " ok\n";
}

print "Checking for XML::SAX.........";
eval {
    require XML::SAX;
    XML::SAX->VERSION($xmlSaxVersion);
};
if ($@) {
    print " failed\n";
    $missing_modules++;
    print <<EOT;
$@
The XML::SAX module version $xmlSaxVersion must be installed.

EOT
    sleep(2);  # Don't hurry too much
} else {
    print " ok\n";
}

print "Checking for XML::LibXML.........";
eval {
    require XML::LibXML;
    XML::LibXML->VERSION($xmlLibXMLVersion);
};
if ($@) {
    print " failed\n";
    $missing_modules++;
    print <<EOT;
$@
The XML::LibXML module version $xmlLibXMLVersion must be installed.

EOT
    sleep(2);  # Don't hurry too much
} else {
    print " ok\n";
}

if ($missing_modules)
{
  print <<EOT;
The missing modules can be obtained from CPAN.  Visit
<URL:http://www.perl.com/CPAN/> to find a CPAN site near you.
For those not on CPAN see the modules directory in this package.

EOT
  exit(1);
}

use ExtUtils::MakeMaker;
# See lib/ExtUtils/MakeMaker.pm for details of how to influence
# the contents of the Makefile that is written.

WriteMakefile(
    'NAME'      => $NAME,
    'VERSION_FROM'   => "PCXFireWall/VeryTight2.pm",
    'DISTNAME'  => $DISTNAME,
    'PREREQ_PM' => { 'PCXFireWall::Config' => $pcx_firewallVersion, 'File::Temp' => $fileTempVersion,
                     'XML::SAX' => $xmlSaxVersion, 'XML::LibXML' => $xmlLibXMLVersion },
    'AUTHOR'    => 'James A. Pattie <james@pcxperience.com>',
    'ABSTRACT'  => 'The PCX FireWall Rules Perl distribution created by Xperience, Inc.',
);
