# VeryTight2.pm - Rules file for generator (IPTables Firewall Configuration).
# Created by James A. Pattie.  Copyright (c) 2001-2004, Xperience, Inc.

package PCXFireWall::VeryTight2;

use strict;
use PCXFireWall::ConfigParser;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.5";

# ****************** User Configuration Starts here ***************************
#  Only edit between "Define your rules here" and "Stop Defining your rules"
# *****************************************************************************

# void startingRules(configObj, interfaceObj, filterObj, natObj, mangleObj, proc, modulesObj, networkObj)
#   Defines the rules/actions that need to be done when the firewall is starting up.
sub startingRules
{
  my $self = shift;
  my %args = ( configObj => undef, interfaceObj => undef, filterObj => undef, natObj => undef, mangleObj => undef, proc => undef, modulesObj => undef, networkObj => undef, @_ );
  my $config = $args{configObj};
  my $filter = $args{filterObj};
  my $nat = $args{natObj};
  my $mangle = $args{mangleObj};
  my $interfaces = $args{interfaceObj};
  my $proc = $args{proc};
  my $modulesObj = $args{modulesObj};
  my $networkObj = $args{networkObj};

  # pull in the configuration values from the XML File.
  my $confObj = PCXFireWall::ConfigParser->new;
  my $dataObj = undef;
  my %data = ();
  eval { $dataObj = $confObj->parse(file => $self->{configFile}, module => "VeryTight2", moduleVersion => $self->{configFileVersion}); };
  if ($@)
  {
    die "Error parsing XML Config File '$self->{configFile}'!\n$@\n";
  }
  %data = %{$dataObj->{data}};

  # any variables I may need to represent the world at large or the firewall machine itself.
  my $localHost = "127.0.0.1";
  my @everyone = ("0.0.0.0/0");

  my @chains = ();  # used in the loops where I walk over chains.
  my $enableForward = ($data{config}{zones}{internal} || $data{config}{zones}{dmz} || $data{config}{zones}{dialin} || $data{config}{zones}{ipsec} ? 1 : 0);

  # The following are log limit rates, etc.
  my $logLevel = $data{config}{logInfo}{level};
  my $taintLogLimit = $data{config}{limit}{log}{taint};
  my $taintLogRate = $data{config}{limitRates}{log}{taint};
  my $reservedLogLimit = $data{config}{limit}{log}{reserved};
  my $reservedLogRate = $data{config}{limitRates}{log}{reserved};
  my $policyLogLimit = $data{config}{limit}{log}{policy};
  my $policyLogRate = $data{config}{limitRates}{log}{policy};
  my $rejectLogLimit = $data{config}{limit}{log}{reject};
  my $rejectLogRate = $data{config}{limitRates}{log}{reject};
  my $normalLogLimit = $data{config}{limit}{log}{normal};
  my $normalLogRate = $data{config}{limitRates}{log}{normal};
  my $icmpLimit = $data{config}{limit}{icmp};
  my $icmpLimitRate = $data{config}{limitRates}{icmp};
  my $bridgeSupport = $data{config}{bridgeSupport};
  my $snortInlineSupport = $data{config}{snortInlineSupport};

  # -------------------- Define your rules here -------------------------------

  # specify the netfilter modules needed.
  if ($data{config}{moduleSupport}{modular})
  {
    $modulesObj->kernelType(type => "modular");
    $modulesObj->register(module => "ip_conntrack");
    $modulesObj->register(module => "iptable_nat");
    $modulesObj->register(module => "iptable_mangle");
    $modulesObj->register(module => "ipt_limit");
    $modulesObj->register(module => "ipt_mark");
    $modulesObj->register(module => "ipt_MARK");
    $modulesObj->register(module => "ipt_MASQUERADE");
    $modulesObj->register(module => "ipt_REDIRECT");
    $modulesObj->register(module => "ipt_REJECT");
    $modulesObj->register(module => "ipt_state");
    $modulesObj->register(module => "ipt_tos");
    $modulesObj->register(module => "ipt_TOS");
    $modulesObj->register(module => "ip_queue");
    $modulesObj->register(module => "ipt_unclean") if ($data{config}{validityChecks}{unclean});
    $modulesObj->register(module => "ipt_multiport");
    
    # handle the conntrackers the user specified.
    foreach my $conntracker (keys %{$data{config}{moduleSupport}{conntrackers}})
    {
      if ($data{config}{moduleSupport}{conntrackers}{$conntracker}{active})
      {
        my $params = $data{config}{moduleSupport}{conntrackers}{$conntracker}{params};
        $modulesObj->register(module => "ip_conntrack_$conntracker", params => $params);
        $modulesObj->register(module => "ip_nat_$conntracker", params => $params);
      }
    }
    
    # handle any other modules the user would like loaded.
    foreach my $module (@{$data{config}{moduleSupport}{modules}})
    {
      $modulesObj->register(module => $module);
    }
  }

  # change log-prefix value
  $config->logPrefix($data{config}{logInfo}{prefix});  # This is the value used if you don't specify the log-prefix value.

  # set default post-comment to be a \n so we have some spacing.
  $config->postComment("\n");
  $config->preComment("");

  # Uncomment the following line if you want to use 'ip' instead of 'ifconfig'.
  $config->networkCommand($data{config}{networkCommand});

  # ECN support
  $proc->sysNetIPv4(file => "tcp_ecn", value => "$data{config}{ecn}");

  # Stop responding to broadcast pings
  $proc->sysNetIPv4(file => "icmp_echo_ignore_broadcasts", value => "1", module => "Nat");

  # turn on synCookie and Martian logging support
  $proc->sysNetIPv4(file => "tcp_syncookies", value => "1");
  $proc->sysNetIPv4Conf(file => "log_martians", value => "1", iface => "all");

  # turn on/off ip forwarding, dynamic IP Addresses, icmp redirects, anti-spoofing blocks and source routing
  if ($enableForward && !$bridgeSupport)
  {
    $proc->sysNetIPv4(file => "ip_forward", value => "1", module => "Nat", required => "true");  # when => "Post",  - let it happen sooner so DNS works 100%.
  }
  $proc->sysNetIPv4(file => "ip_dynaddr", value => "1", module => "Nat");
  $proc->sysNetIPv4Conf(file => "rp_filter", value => "1", module => "Nat", iface => "all");
  if ($data{config}{zones}{ipsec})
  {
    my %processedInterfaces = ();  # used to keep duplicate entries from being generated (not really necessary just nicer looking).
    foreach my $alias (keys %{$data{config}{interfaces}{ipsec}})
    {
      my $disabled = 0;
      foreach my $zone ("external", "internal", "dmz", "dialin")
      {
        next if ($zone !~ /^(external)$/ && !$data{config}{zones}{$zone});
        if (exists $data{config}{interfaces}{$zone}{$alias})
        {
          $disabled = !$data{config}{interfaces}{$zone}{$alias}{active};
          last;
        }
      }
      next if ($disabled);

      my $externalInterface = $data{config}{interfaces}{aliasLookup}{$data{config}{interfaces}{ipsec}{$alias}->{boundTo}};
      if (!exists $processedInterfaces{$externalInterface})
      {
        $proc->sysNetIPv4Conf(file => "rp_filter", value => "0", module => "Nat", iface => "$externalInterface");  # turn off anti-spoofing on VPN interface
        $processedInterfaces{$externalInterface} = 1;
      }
    }
  }
  $proc->sysNetIPv4Conf(file => "accept_source_route", value => "0", module => "Nat", iface => "all");
  $proc->sysNetIPv4Conf(file => "accept_redirects", value => "0", module => "Nat", iface => "all");

  # create the interface definitions and enable proxy_arp as needed.
  foreach my $zone ("external", "internal", "ipsec", "dmz", "dialin")
  {
    if ($zone =~ /^(ipsec|dialin)$/)
    {
      # create the zone location, but only if we are using this zone. (dmz is already builtin to the Toolkit)
      if ($data{config}{zones}{$zone})
      {
        $interfaces->defineLocation(location => $zone);
      }
    }

    # for each zone, create the interfaces defined in it, but only if the zone is active.
    if ($zone =~ /^(external)$/ || $data{config}{zones}{$zone})
    {
      foreach my $alias (keys %{$data{config}{interfaces}{$zone}})
      {
        next if (!$data{config}{interfaces}{$zone}{$alias}{active});

        my $interface = $data{config}{interfaces}{$zone}{$alias}{name};
        my $interfaceName = $interface;
        my $ip = "";
        my $type = "static";
        my $proxyarp = $data{config}{interfaces}{$zone}{$alias}{proxyarp};
        my $dynamicIPMode = $data{config}{dynamicInterfacesMode};
        if ($zone !~ /^(ipsec)$/)
        {
          if ($data{config}{interfaces}{$zone}{$alias}->{ip} eq "dynamic")
          {
            $type = "dynamic";
          }
          else
          {
            $ip = $data{config}{interfaces}{$zone}{$alias}->{ip};
          }
        }
        else
        {
          # find the IP Address of the interface that we are boundTo.
          foreach my $tmpZone ("external", "internal", "dmz", "dialin")
          {
            my $tmpAlias = $data{config}{interfaces}{$zone}{$alias}->{boundTo};
            if (exists $data{config}{interfaces}{$tmpZone}{$tmpAlias})
            {
              if ($data{config}{interfaces}{$tmpZone}{$tmpAlias}->{ip} eq "dynamic")
              {
                $type = "dynamic";
              }
              else
              {
                $ip = $data{config}{interfaces}{$tmpZone}{$tmpAlias}->{ip};
              }
            }
          }
        }
        $interfaces->defineInterface(name => $interface, location => $zone, type => $type, IP => $ip, alias => $alias, dynamicIPMode => $dynamicIPMode);

        if ($proxyarp)
        {
          # enable proxy_arp support for this interface.
          $proc->sysNetIPv4Conf(file => "proxy_arp", value => "1", module => "Nat", iface => "$interface", when => "Post");
        }
      }
    }
  }
  # define the loopback interface
  $interfaces->defineInterface(name => "lo", location => "firewall", type => "static", IP => "127.0.0.1", alias => "loopback");

  # set chain policies
  $filter->policy(chain => "FORWARD", policy => "DROP");
  $filter->policy(chain => "INPUT", policy => "DROP");
  $filter->policy(chain => "OUTPUT", policy => "DROP");

  $nat->policy(chain => "PREROUTING", policy => "DROP");
  $nat->policy(chain => "POSTROUTING", policy => "DROP");
  $nat->policy(chain => "OUTPUT", policy => "DROP");

  $mangle->policy(chain => "OUTPUT", policy => "ACCEPT");
  $mangle->policy(chain => "PREROUTING", policy => "ACCEPT");
  $mangle->policy(chain => "INPUT", policy => "ACCEPT");
  $mangle->policy(chain => "FORWARD", policy => "ACCEPT");
  $mangle->policy(chain => "POSTROUTING", policy => "ACCEPT");

  # flush and delete all chains
  $filter->deleteChain(name => "All");
  $mangle->deleteChain(name => "All");
  $nat->deleteChain(name => "All");

  # setup chains to log then drop packets.
  $filter->createChain(name => "Tainted") if ($data{config}{validityChecks}{unclean});
  $filter->createChain(name => "tcpFlags") if ($data{config}{validityChecks}{tcpFlags});
  #$nat->createChain(name => "Tainted");
  $nat->createChain(name => "Reserved");

  if ($data{config}{validityChecks}{unclean})
  {
    $filter->log(chain => "Tainted", ruleMode => "setup", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Tainted Packet: ", "log-level" => "$logLevel");
    $filter->drop(chain => "Tainted", ruleMode => "setup",);

    #$nat->log(chain => "Tainted", ruleMode => "setup", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid Packet: ", "log-level" => "$logLevel");
    #$nat->drop(chain => "Tainted", ruleMode => "setup",);
  }

  $nat->log(chain => "Reserved", ruleMode => "setup", "log-limit" => $reservedLogLimit, "log-limit-rate" => $reservedLogRate, "log-prefix" => "Reserved Packet: ", "log-level" => "$logLevel");
  $nat->drop(chain => "Reserved", ruleMode => "setup",);

  # setup tcp flag checks.
  if ($data{config}{validityChecks}{tcpFlags})
  {
    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ALL NONE", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ALL NONE", preComment => "# All bits are cleared.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "SYN,FIN SYN,FIN", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "SYN,FIN SYN,FIN", preComment => "# SYN and FIN are both set.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "SYN,RST SYN,RST", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "SYN,RST SYN,RST", preComment => "# SYN and RST are both set.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "FIN,RST FIN,RST", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "FIN,RST FIN,RST", preComment => "# FIN and RST are both set.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,FIN FIN", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,FIN FIN", preComment => "# FIN is set without ACK.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,PSH PSH", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,PSH PSH", preComment => "# PSH is set without ACK.\n");

    $filter->log(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,URG URG", "log-limit" => $taintLogLimit, "log-limit-rate" => $taintLogRate, "log-prefix" => "Invalid TCP Flags: ", "log-level" => "$logLevel");
    $filter->drop(chain => "tcpFlags", ruleMode => "setup", p => "tcp", "tcp-flags" => "ACK,URG URG", preComment => "# URG is set without ACK.\n");

    # check for invalid packets.
    $filter->jump(chain => "INPUT", dchain => "tcpFlags", p => "tcp");
    if ($enableForward || $bridgeSupport)
    {
      $filter->jump(chain => "FORWARD", dchain => "tcpFlags", p => "tcp");
    }
  }

  foreach my $alias (keys %{$data{config}{interfaces}{external}})
  {
    next if (!$data{config}{interfaces}{external}{$alias}{active});

    my $externalInterface = $data{config}{interfaces}{aliasLookup}{$alias};
    next if ($externalInterface =~ /:/);  # don't do anything if it is an alias interface.
    # handle unclean packets coming in.
    if ($data{config}{validityChecks}{unclean})
    {
      $filter->jump(chain => "INPUT", dchain => "Tainted", i => $externalInterface, unclean => 1);
      #$nat->jump(chain => "PREROUTING", dchain => "Tainted", i => $externalInterface, unclean => 1);
    }

    # block reserved/private addresses coming in the external interface.
    $nat->jump(chain => "PREROUTING", dchain => "Reserved", i => $externalInterface, s => "inInterfaceIP", preComment => "# Block Reserved/Private Addresses coming in the external interface.\n") if ($data{config}{interfaces}{external}{$alias}{type} eq "static" || $data{config}{dynamicInterfacesMode} eq "lookupIP");

    foreach my $address (@{$data{config}{reserved}})
    {
      next if (!$address->{active});

      my $found = 0;
      foreach my $limitTo (@{$address->{limitTo}})
      {
        if ($alias eq $limitTo)
        {
          $found = 1;
          last;
        }
      }
      next if (!$found && scalar @{$address->{limitTo}} > 0);

      $nat->jump(chain => "PREROUTING", dchain => "Reserved", i => $externalInterface, s => $address->{host}, preComment => "# Block Reserved/Private Addresses coming in the external interface.\n");
    }
  }

  # now define all the networkObj Network entries
  foreach my $index (sort { $a <=> $b } keys %{$data{config}{networks}})
  {
    eval { $networkObj->register(index => $index, host => $data{config}{networks}{$index}{host},
                          comment => $data{config}{networks}{$index}{comment}); };
    if ($@)
    {
      die "Failed to register networks[$index]='$data{config}{networks}{$index}{host}'!\n";
    }
  }

  # setup the user customizable chains to support installations that need to dynamically add/remove rules from the firewall.
  $filter->createChain(name => "RemoteAccess");
  $filter->createChain(name => "RemoteAccessOUT");
  if ($enableForward || $bridgeSupport)
  {
    $filter->createChain(name => "RemoteAccessForward");
  }
  $nat->createChain(name => "RemoteAccess");
  $nat->createChain(name => "RemoteAccessPOST");
  $nat->createChain(name => "RemoteAccessOUT");

  # setup the chains that we will jump to for incoming tcp, udp, icmp, other packets (destined for the box)
  $filter->createChain(name => "incomingLOCAL");
  $filter->createChain(name => "incomingTCP");
  $filter->createChain(name => "incomingUDP");
  $filter->createChain(name => "incomingICMP");
  $filter->createChain(name => "incomingOTHER");

  $nat->createChain(name => "preLOCAL");
  $nat->createChain(name => "preTCP");
  $nat->createChain(name => "preUDP");
  $nat->createChain(name => "preICMP");
  $nat->createChain(name => "preOTHER");

  $nat->createChain(name => "outgoingLOCAL");
  $nat->createChain(name => "outgoingTCP");
  $nat->createChain(name => "outgoingUDP");
  $nat->createChain(name => "outgoingICMP");
  $nat->createChain(name => "outgoingOTHER");

  $nat->createChain(name => "postLOCAL");
  $nat->createChain(name => "postTCP");
  $nat->createChain(name => "postUDP");
  $nat->createChain(name => "postICMP");
  $nat->createChain(name => "postOTHER");

  # setup the chains that we will jump to for outgoing tcp, udp, icmp, other packets (locally generated)
  $filter->createChain(name => "outgoingLOCAL");
  $filter->createChain(name => "outgoingTCP");
  $filter->createChain(name => "outgoingUDP");
  $filter->createChain(name => "outgoingICMP");
  $filter->createChain(name => "outgoingOTHER");

  if ($enableForward || $bridgeSupport)
  {
    # setup the chains that we will jump to for forwarded tcp, udp, icmp, other packets
    $filter->createChain(name => "forwardedTCP");
    $filter->createChain(name => "forwardedUDP");
    $filter->createChain(name => "forwardedICMP");
    $filter->createChain(name => "forwardedOTHER");
  }

  # jump into the user customizable chains first
  $filter->jump(chain => "INPUT", dchain => "RemoteAccess");
  $filter->jump(chain => "OUTPUT", dchain => "RemoteAccessOUT");
  if ($enableForward || $bridgeSupport)
  {
    $filter->jump(chain => "FORWARD", dchain => "RemoteAccessForward");
  }
  $nat->jump(chain => "PREROUTING", dchain => "RemoteAccess");
  $nat->jump(chain => "POSTROUTING", dchain => "RemoteAccessPOST");
  $nat->jump(chain => "OUTPUT", dchain => "RemoteAccessOUT");

  # now do statefull traffic.
  my @chains = ("INPUT", "OUTPUT");
  if ($enableForward || $bridgeSupport)
  {
    push @chains, "FORWARD";
  }
  foreach my $chain (@chains)
  {
    # we QUEUE the traffic if snortInlineSupport is enabled and this is the FORWARD chain or we only have the external zone defined.
    if (($chain eq "FORWARD" || !$enableForward) && $snortInlineSupport)
    {
      $filter->queue(chain => "$chain", state => 1, states => "ESTABLISHED,RELATED", ruleMode => "setup", preComment => "# QUEUE all traffic that is already established and/or related for processing by snort-inline.\n");
    }
    else
    {
      $filter->accept(chain => "$chain", state => 1, states => "ESTABLISHED,RELATED", ruleMode => "setup", preComment => "# Accept all traffic that is already established and/or related.\n");
    }
  }

  # Do any NAT necessary, but only if some other zone than external is enabled.
  if ($enableForward && !$bridgeSupport)
  {
    foreach my $alias (keys %{$data{config}{interfaces}{external}})
    {
      next if (!$data{config}{interfaces}{external}{$alias}{active});

      my $externalInterface = $data{config}{interfaces}{aliasLookup}{$alias};
      next if ($externalInterface =~ /:/);  # don't do anything if it is an alias interface.
      if ($data{config}{interfaces}{external}{$alias}{ip} eq "dynamic")
      {
        $nat->masquerade(chain => "POSTROUTING", ruleMode => "setup", o => $externalInterface, preComment => "# MASQUERADE any traffic going out $externalInterface to come from this box.\n\n");
      }
      else
      {
        $nat->snat(chain => "POSTROUTING", ruleMode => "setup", o => $externalInterface, "to-source" => "outInterfaceIP", preComment => "# SNAT any traffic going out $externalInterface to come from this box.\n\n");
      }
    }
  }

  # allow all incoming traffic that is originating from the box itself.
  $filter->jump(chain => "INPUT", dchain => "incomingLOCAL", i => "lo");
  $filter->accept(chain => "INPUT", ruleMode => "setup", i => "lo", preComment => "# Accept all Locally Generated Traffic.\n");

  # allow all outgoing traffic that is originating from the box to itself.
  $filter->jump(chain => "OUTPUT", dchain => "outgoingLOCAL", o => "lo");
  $filter->accept(chain => "OUTPUT", ruleMode => "setup", o => "lo", preComment => "# Accept all Locally Generated Traffic going to ourself.\n");

  $nat->jump(chain => "POSTROUTING", dchain => "postLOCAL", o => "lo");
  $nat->accept(chain => "POSTROUTING", ruleMode => "setup", o => "lo", preComment => "# Accept all Locally Generated Traffic going to ourself.\n");

  $nat->jump(chain => "OUTPUT", dchain => "outgoingLOCAL", o => "lo");
  $nat->accept(chain => "OUTPUT", ruleMode => "setup", o => "lo", preComment => "# Accept all Locally Generated Traffic going to ourself.\n");

  $nat->jump(chain => "PREROUTING", dchain => "preLOCAL", i => "lo");
  $nat->accept(chain => "PREROUTING", ruleMode => "setup", i => "lo", preComment => "# Accept all Locally Generated Traffic going to ourself.\n");

  # now jump to the specified chains based upon what is going on
  $filter->jump(chain => "INPUT", dchain => "incomingTCP", p => "tcp");
  $filter->jump(chain => "INPUT", dchain => "incomingUDP", p => "udp");
  $filter->jump(chain => "INPUT", dchain => "incomingICMP", p => "icmp");
  $filter->jump(chain => "INPUT", dchain => "incomingOTHER");

  $filter->jump(chain => "OUTPUT", dchain => "outgoingTCP", p => "tcp");
  $filter->jump(chain => "OUTPUT", dchain => "outgoingUDP", p => "udp");
  $filter->jump(chain => "OUTPUT", dchain => "outgoingICMP", p => "icmp");
  $filter->jump(chain => "OUTPUT", dchain => "outgoingOTHER");

  if ($enableForward || $bridgeSupport)
  {
    $filter->jump(chain => "FORWARD", dchain => "forwardedTCP", p => "tcp");
    $filter->jump(chain => "FORWARD", dchain => "forwardedUDP", p => "udp");
    $filter->jump(chain => "FORWARD", dchain => "forwardedICMP", p => "icmp");
    $filter->jump(chain => "FORWARD", dchain => "forwardedOTHER");
  }

  $nat->jump(chain => "PREROUTING", dchain => "preTCP", p => "tcp");
  $nat->jump(chain => "PREROUTING", dchain => "preUDP", p => "udp");
  $nat->jump(chain => "PREROUTING", dchain => "preICMP", p => "icmp");
  $nat->jump(chain => "PREROUTING", dchain => "preOTHER");

  $nat->jump(chain => "POSTROUTING", dchain => "postTCP", p => "tcp");
  $nat->jump(chain => "POSTROUTING", dchain => "postUDP", p => "udp");
  $nat->jump(chain => "POSTROUTING", dchain => "postICMP", p => "icmp");
  $nat->jump(chain => "POSTROUTING", dchain => "postOTHER");

  $nat->jump(chain => "OUTPUT", dchain => "outgoingTCP", p => "tcp");
  $nat->jump(chain => "OUTPUT", dchain => "outgoingUDP", p => "udp");
  $nat->jump(chain => "OUTPUT", dchain => "outgoingICMP", p => "icmp");
  $nat->jump(chain => "OUTPUT", dchain => "outgoingOTHER");

  # add the tcp, udp, icmp checks in the OTHER chains so that we don't accidentally match traffic that is destined to be blocked.
  @chains = ("incomingOTHER", "outgoingOTHER");
  if ($enableForward || $bridgeSupport)
  {
    push @chains, "forwardedOTHER";
  }
  foreach my $chain (@chains)
  {
    foreach my $protocol ("tcp", "udp", "icmp")
    {
      $filter->return(chain => $chain, p => $protocol);
    }
  }

  foreach my $chain ("preOTHER", "postOTHER", "outgoingOTHER")
  {
    foreach my $protocol ("tcp", "udp", "icmp")
    {
      $nat->return(chain => $chain, p => $protocol);
    }
  }

  # now do default logging.
  @chains = ("INPUT", "OUTPUT");
  if ($enableForward || $bridgeSupport)
  {
    push @chains, "FORWARD";
  }
  foreach my $chain (@chains)
  {
    $filter->log(chain => "$chain", ruleMode => "setup", "log-prefix" => "FILTER DROP - $chain:", "log-limit" => $policyLogLimit, "log-limit-rate" => $policyLogRate, preComment => "# LOG all packets that are about to be dropped because of the POLICY.\n", "log-level" => "$logLevel");
  }

  foreach my $chain ("PREROUTING", "POSTROUTING", "OUTPUT")
  {
    $nat->log(chain => "$chain", ruleMode => "setup", "log-prefix" => "NAT DROP - $chain:", "log-limit" => $policyLogLimit, "log-limit-rate" => $policyLogRate, preComment => "# LOG all packets that are about to be dropped because of the POLICY.\n", "log-level" => "$logLevel");
  }

  #*********** Everything after this must be in a protocol specific chain for filter otherwise you will probably not get processed correctly
  # and you will see an entry logged! *************************************************************************************************************

  # setup IPSec support if needed.
  if ($data{config}{zones}{ipsec})
  {
    foreach my $zone ("external", "internal", "dmz", "dialin")
    {
      if ($zone =~ /^(external)$/ || $data{config}{zones}{$zone})
      {
        foreach my $alias (keys %{$data{config}{interfaces}{$zone}})
        {
          next if (!$data{config}{interfaces}{$zone}{$alias}{active});

          if ($data{config}{interfaces}{$zone}{$alias}{ipsec})
          {
            my $interface = $data{config}{interfaces}{$zone}{$alias}{name};
            my $ipsecLimitFrom = $data{config}{interfaces}{$zone}{$alias}{ipsecLimitFrom};
            foreach my $sourceIndex ((scalar @{$ipsecLimitFrom} > 0 ? @{$ipsecLimitFrom} : (($zone !~ /^(external)$/ && (scalar @{$data{config}{interfaces}{$zone}{$alias}{network}} > 0)) ? @{$data{config}{interfaces}{$zone}{$alias}{network}} : "")))
            {
              my $source = "";
              $source = "\${NETWORK_$sourceIndex}" if (length $sourceIndex > 0 && $sourceIndex >= 0);
              next if ($sourceIndex >= 0 && !$data{config}{networks}{$sourceIndex}{active});

              # allow UDP Port 500
              $filter->accept(chain => "incomingUDP", i => $interface, s => $source, d => "inInterfaceIP", p => "udp", dport => "500", sport => "500", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $filter->accept(chain => "outgoingUDP", o => $interface, s => "outInterfaceIP", d => $source, p => "udp", dport => "500", sport => "500", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "preUDP", i => $interface, s => $source, d => "inInterfaceIP", p => "udp", dport => "500", sport => "500", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $nat->accept(chain => "postUDP", o => $interface, s => "outInterfaceIP", d => $source, p => "udp", dport => "500", sport => "500", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "outgoingUDP", o => $interface, s => "outInterfaceIP", d => $source, p => "udp", dport => "500", sport => "500", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");

              # allow Protocols 50 & 51
              $filter->accept(chain => "incomingOTHER", i => $interface, s => $source, d => "inInterfaceIP", p => "50", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $filter->accept(chain => "outgoingOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "50", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "preOTHER", i => $interface, s => $source, d => "inInterfaceIP", p => "50", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $nat->accept(chain => "postOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "50", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "outgoingOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "50", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");

              $filter->accept(chain => "incomingOTHER", i => $interface, s => $source, d => "inInterfaceIP", p => "51", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $filter->accept(chain => "outgoingOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "51", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "preOTHER", i => $interface, s => $source, d => "inInterfaceIP", p => "51", preComment => "# Allow VPN traffic to get to us, zone = $zone.\n");
              $nat->accept(chain => "postOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "51", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
              $nat->accept(chain => "outgoingOTHER", o => $interface, s => "outInterfaceIP", d => $source, p => "51", preComment => "# Allow VPN traffic to get out from us, zone = $zone.\n");
            }
          }
        }
      }
    }
  }

  # now we go into a loop that will walk over the services hash and for each service determine what needs to be done.
  foreach my $service (@{$data{paths}{serviceGroupsOrder}})
  {
    next if (!$data{paths}{$service}{active});
    # get a hash of the zonePaths for this service.
    my %zonePaths = %{$data{paths}{$service}};
    foreach my $path (@{$zonePaths{zonePathOrder}})
    {
      # figure out what zone this path is part of.
      my $fromZone = $path;
      my $toZone = $path;
      if ($fromZone =~ /Broadcast$/)
      {
        $fromZone =~ s/^((.*)(Broadcast))$/$2/;
        $toZone = "Broadcast";
      }
      if ($fromZone =~ /To/)
      {
        $fromZone =~ s/^((.*)(To.*))$/$2/;
        $toZone =~ s/^((.*To)(.*))$/$3/;
      }
      $fromZone = lc($fromZone);  # make sure the zone is lowercase.
      $toZone = lc($toZone);

      # don't do anything with this path if it is working with a zone we don't have enabled.
      next if ($fromZone =~ /^(internal|ipsec|dmz|dialin)$/ && !$data{config}{zones}{$fromZone});
      next if ($toZone =~ /^(internal|ipsec|dmz|dialin)$/ && !$data{config}{zones}{$toZone});

      if ($fromZone !~ /^(external|internal|ipsec|dmz|dialin|firewall)$/)
      {
        die "Could not correctly determine fromZone from path = '$path', fromZone = '$fromZone'!\n";
      }
      if (length $toZone && $toZone !~ /^(external|internal|ipsec|dmz|dialin|firewall|broadcast)$/)
      {
        die "Could not correctly determine toZone from path = '$path', fromZone = '$toZone'!\n";
      }

      #print "Path : $path, fromZone = $fromZone, toZone = $toZone\n";

      # handle paths that only have i and limitFrom attributes and support all actions.
      if ($path =~ /^(externalToFirewall|internalToFirewall|ipsecToFirewall|dmzToFirewall|dialinToFirewall|externalBroadcast|internalBroadcast|ipsecBroadcast|dmzBroadcast|dialinBroadcast)$/)
      {
        foreach my $pathNode (@{$zonePaths{$path}})
        {
          next if (!$pathNode->{active});
          my $i = $pathNode->{i};
          my $limitFrom = $pathNode->{limitFrom};
          my $reject = $pathNode->{reject};
          my $queue = $pathNode->{queue};
          my $redirect = $pathNode->{redirect};
          my $dnat = $pathNode->{dnat};
          my $dnatDestinationId = $pathNode->{dnatDestination};
          my $dnatDestination = "";
          $dnatDestination = "\${NETWORK_$dnatDestinationId}" if ($dnatDestinationId);
          my $dnatZone = $pathNode->{dnatZone};
          next if ($dnatZone =~ /^(internal|ipsec|dmz|dialin)$/ && !$data{config}{zones}{$dnatZone});
          my $dnatInterface = $data{config}{interfaces}{aliasLookup}{$pathNode->{dnatInterface}};
          my $dropNoLog = $pathNode->{dropNoLog};
          my $log = $pathNode->{log};
          my $logComment = $pathNode->{logComment};
          my $logLevel = $pathNode->{logLevel};
          my @infos = $dataObj->returnPathServiceInfos($pathNode->{service});
          my $dest = ($toZone eq "broadcast" ? "inInterfaceBroadcast" : ($pathNode->{ip} eq "" ? "inInterfaceIP" : "\${NETWORK_$pathNode->{ip}}"));

          foreach my $alias (($i->[0] ne "*" ? @{$i} : (keys %{$data{config}{interfaces}{$fromZone}})))
          {
            next if (!$data{config}{interfaces}{$fromZone}{$alias}{active});

            my $interface = $data{config}{interfaces}{aliasLookup}{$alias};
            foreach my $sourceIndex ((scalar @{$limitFrom} > 0 ? @{$limitFrom} : (($fromZone !~ /^(external|firewall)$/ && (scalar @{$data{config}{interfaces}{$fromZone}{$alias}{network}} > 0)) ? @{$data{config}{interfaces}{$fromZone}{$alias}{network}} : "")))
            {
              my $source = "";
              my %macInfo = ();
              if (length $sourceIndex > 0 && $sourceIndex >= 0)
              {
                next if (!$data{config}{networks}{$sourceIndex}{active});

                if ($data{config}{networks}{$sourceIndex}{type} eq "normal")
                {
                  $source = "\${NETWORK_$sourceIndex}";
                }
                elsif ($data{config}{networks}{$sourceIndex}{type} eq "mac")
                {
                  $macInfo{mac} = 1;
                  $macInfo{'mac-source'} = "\${NETWORK_$sourceIndex}";
                }
              }

              my %logInfo = ();
              if ($log)
              {
                %logInfo = ("log" => 1, "log-prefix" => $logComment, "log-limit" => $normalLogLimit, "log-limit-rate" => $normalLogRate, "log-level" => $logLevel);
              }
              if ($reject)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, %macInfo);

                    $filter->reject(chain => "incoming" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, "reject-with" => ($proto eq "tcp" ? "tcp-reset" : ""), %logInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, %macInfo);

                      $filter->reject(chain => "incoming" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, "reject-with" => ($proto eq "tcp" ? "tcp-reset" : ""), %logInfo, %macInfo);
                    }
                  }
                }
              }
              elsif ($queue)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %macInfo);

                    $filter->queue(chain => "incoming" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %logInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %macInfo);

                      $filter->queue(chain => "incoming" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %logInfo, %macInfo);
                    }
                  }
                }
              }
              elsif ($redirect)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};
                  my $redirectPort = (length $pathNode->{redirectPort} > 0 ? $pathNode->{redirectPort} : $dport);

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->redirect(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Redirected\n", "to-ports" => $redirectPort, %icmpInfo, %logInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->redirect(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - Redirected\n", "to-ports" => $redirectPort, %icmpInfo, %logInfo, %macInfo);
                    }

                    $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$redirectPort",
                                preComment => "# Service $service, path $path - Redirected\n", %icmpInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$redirectPort",
                                  preComment => "# Service $service, path $path - Redirected\n", %icmpInfo, %macInfo);
                    }
                  }
                }
              }
              elsif ($dropNoLog)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->drop(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %macInfo);


                    $filter->drop(chain => "incoming" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %logInfo, %macInfo);
                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->drop(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %macInfo);

                      $filter->drop(chain => "incoming" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %logInfo, %macInfo);
                    }
                  }
                }
              }
              elsif ($dnat)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};
                  my $dnatPort = (length $pathNode->{dnatPort} > 0 ? $pathNode->{dnatPort} : $dport);

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->dnat(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Dnated\n", "to-destination" => "$dnatDestination" . ($proto =~ /^(tcp|udp)$/ && length $dnatPort > 0 ? ":$dnatPort" : ""), %icmpInfo, %logInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->dnat(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path - Dnated\n", "to-destination" => "$dnatDestination" . ($proto =~ /^(tcp|udp)$/ && length $dnatPort > 0 ? ":$dnatPort" : ""), %icmpInfo, %logInfo, %macInfo);
                    }

                    # now we have to potentially allow it through the firewall
                    if ($dest =~ /^(inInterfaceBroadcast|inInterfaceIP)$/)
                    {
                      if ($fromZone ne $dnatZone)
                      {
                        $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, d => $dnatDestination, o => $dnatInterface,
                                  s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                  preComment => "# Service $service, path $path - Dnated\n", %icmpInfo, %macInfo);

                        # now we have to allow it out of the firewall (if the fromZone and dnatZone are the same then we have to snat.
                        $nat->accept(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                    preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                      }
                      else
                      {
                        $nat->snat(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                    preComment => "# Service $service, path $path - Dnated into zone $dnatZone\n", "to-source" => "outInterfaceIP", %icmpInfo);
                      }
                    }
                    else
                    {
                      $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, d => $dnatDestination, o => $dnatInterface,
                                s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                preComment => "# Service $service, path $path - Dnated\n", %icmpInfo, %macInfo);

                      if ($fromZone ne $dnatZone)
                      {
                        # now we have to allow it out of the firewall (if the fromZone and dnatZone are the same then we have to snat.
                        $nat->accept(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                    preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                      }
                      else
                      {
                        $nat->snat(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                    preComment => "# Service $service, path $path - Dnated into zone $dnatZone\n", "to-source" => "outInterfaceIP", %icmpInfo);
                      }
                    }
                  }
                }
              }
              else  # handle as a normal ACCEPT.
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path\n", %icmpInfo, %macInfo);

                    $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => $dest,
                                s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path\n", %icmpInfo, %logInfo, %macInfo);

                    if ($toZone eq "broadcast") # handle the global broadcast address of 255.255.255.255
                    {
                      $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path\n", %icmpInfo, %macInfo);

                      $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => "255.255.255.255",
                                  s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                  preComment => "# Service $service, path $path\n", %icmpInfo, %logInfo, %macInfo);
                    }
                  }
                }
              }
            } # end of source
          } # end of interface
        }
      }
      # handle paths with only o and limitTo and do not support all actions.
      elsif ($path =~ /^(firewallTo(Internal|External|IPSec|DMZ|Dialin))$/)
      {
        foreach my $pathNode (@{$zonePaths{$path}})
        {
          next if (!$pathNode->{active});
          my $o = $pathNode->{o};
          my $limitTo = $pathNode->{limitTo};
          my $reject = $pathNode->{reject};
          my $queue = $pathNode->{queue};
          my $redirect = $pathNode->{redirect};
          my $dnat = $pathNode->{dnat};
          my $dnatDestinationId = $pathNode->{dnatDestination};
          my $dnatDestination = "";
          $dnatDestination = "\${NETWORK_$dnatDestinationId}" if ($dnatDestinationId);
          my $dnatZone = $pathNode->{dnatZone};
          next if ($dnatZone =~ /^(internal|ipsec|dmz|dialin)$/ && !$data{config}{zones}{$dnatZone});
          my $dnatInterface = $data{config}{interfaces}{aliasLookup}{$pathNode->{dnatInterface}};
          my $log = $pathNode->{log};
          my $logComment = $pathNode->{logComment};
          my $logLevel = $pathNode->{logLevel};
          my @infos = $dataObj->returnPathServiceInfos($pathNode->{service});

          foreach my $alias (($o->[0] ne "*" ? @{$o} : (keys %{$data{config}{interfaces}{$toZone}})))
          {
            next if (!$data{config}{interfaces}{$toZone}{$alias}{active});

            my $interface = $data{config}{interfaces}{aliasLookup}{$alias};
            foreach my $destIndex ((scalar @{$limitTo} > 0 ? @{$limitTo} : (($toZone !~ /^(external)$/ && (scalar @{$data{config}{interfaces}{$toZone}{$alias}{network}} > 0)) ? @{$data{config}{interfaces}{$toZone}{$alias}{network}} : "")))
            {
              my $dest = "";
              $dest = "\${NETWORK_$destIndex}" if (length $destIndex > 0 && $destIndex >= 0);
              next if (!$data{config}{networks}{$destIndex}{active});

              my %logInfo = ();
              if ($log)
              {
                %logInfo = ("log" => 1, "log-prefix" => $logComment, "log-limit" => $normalLogLimit, "log-limit-rate" => $normalLogRate, "log-level" => $logLevel);
              }
              if ($reject)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Rejected\n", %icmpInfo);

                    $filter->reject(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, "reject-with" => ($proto eq "tcp" ? "tcp-reset" : ""), %logInfo);
                  }
                }
              }
              elsif ($queue)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo);

                    $filter->queue(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %logInfo);
                  }
                }
              }
              elsif ($redirect)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};
                  my $redirectPort = (length $pathNode->{redirectPort} > 0 ? $pathNode->{redirectPort} : $dport);

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->redirect(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Redirected\n", "to-ports" => $redirectPort, %icmpInfo, %logInfo);

                    $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => "inInterfaceIP",
                                s => "inInterfaceIP", p => $proto, sport => "$sport", dport => "$redirectPort",
                                preComment => "# Service $service, path $path - Redirected\n", %icmpInfo);
                  }
                }
              }
              elsif ($dnat)
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};
                  my $dnatPort = (length $pathNode->{dnatPort} > 0 ? $pathNode->{dnatPort} : $dport);

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->dnat(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path - Dnated\n", "to-destination" => "$dnatDestination" . ($proto =~ /^(tcp|udp)$/ && length $dnatPort > 0 ? ":$dnatPort" : ""), %icmpInfo, %logInfo);

                    # now we have to potentially allow it through the firewall
                    if ($toZone ne $dnatZone)
                    {
                      $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, d => $dnatDestination, o => $dnatInterface,
                                s => "inInterfaceIP", p => $proto, sport => "$sport", dport => "$dnatPort",
                                preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                    }
                    else
                    {
                      $filter->accept(chain => "outgoing" . $chainEnding, d => $dnatDestination, o => $dnatInterface,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dnatPort",
                                preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                    }
                    $nat->accept(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dnatPort",
                                preComment => "# Service $service, path $path - Dnated into zone $dnatZone\n", %icmpInfo);
                  }
                }
              }
              else  # handle as a normal ACCEPT.
              {
                foreach my $infoNode (@infos)
                {
                  my $p = $infoNode->{proto};
                  my $dport = $infoNode->{dport};
                  my $sport = $infoNode->{sport};
                  my $icmpType = $infoNode->{icmpType};

                  foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                  {
                    my %icmpInfo = ();
                    if ($proto eq "icmp")
                    {
                      $icmpInfo{limit} = $icmpLimit;
                      $icmpInfo{'limit-rate'} = $icmpLimitRate;
                      $icmpInfo{'icmp-type'} = $icmpType;
                    }

                    my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                    $nat->accept(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path\n", %icmpInfo);

                    $filter->accept(chain => "outgoing" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path\n", %icmpInfo, %logInfo);

                    $nat->accept(chain => "post" . $chainEnding, o => $interface, d => $dest,
                                s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dport",
                                preComment => "# Service $service, path $path\n", %icmpInfo);
                  }
                }
              }
            } # end of dest
          } # end of interface
        }
      }
      # handle paths with i, o, limitFrom, limitTo and support all options.
      elsif ($path =~ /^(internalTo(Internal|External|IPSec|DMZ|Dialin)|externalTo(External|Internal|IPSec|DMZ|Dialin)|ipsecTo(Internal|External|IPSec|DMZ|Dialin)|dmzTo(Internal|External|IPSec|DMZ|Dialin)|dialinTo(Internal|External|IPSec|DMZ|Dialin))$/)
      {
        foreach my $pathNode (@{$zonePaths{$path}})
        {
          next if (!$pathNode->{active});
          my $i = $pathNode->{i};
          my $o = $pathNode->{o};
          my $limitFrom = $pathNode->{limitFrom};
          my $limitTo = $pathNode->{limitTo};
          my $reject = $pathNode->{reject};
          my $queue = $pathNode->{queue};
          my $redirect = $pathNode->{redirect};
          my $dnat = $pathNode->{dnat};
          my $dnatDestinationId = $pathNode->{dnatDestination};
          my $dnatDestination = "";
          $dnatDestination = "\${NETWORK_$dnatDestinationId}" if ($dnatDestinationId);
          my $dnatZone = $pathNode->{dnatZone};
          next if ($dnatZone =~ /^(ipsec|dmz|dialin)$/ && !$data{config}{zones}{$dnatZone});
          my $dnatInterface = $data{config}{interfaces}{aliasLookup}{$pathNode->{dnatInterface}};
          my $dropNoLog = $pathNode->{dropNoLog};
          my $log = $pathNode->{log};
          my $logComment = $pathNode->{logComment};
          my $logLevel = $pathNode->{logLevel};
          my @infos = $dataObj->returnPathServiceInfos($pathNode->{service});

          foreach my $alias (($i->[0] ne "*" ? @{$i} : (keys %{$data{config}{interfaces}{$fromZone}})))
          {
            next if (!$data{config}{interfaces}{$fromZone}{$alias}{active});

            my $interface = $data{config}{interfaces}{aliasLookup}{$alias};
            foreach my $oAlias (($o->[0] ne "*" ? @{$o} : (keys %{$data{config}{interfaces}{$toZone}})))
            {
              next if (!$data{config}{interfaces}{$toZone}{$oAlias}{active});

              my $oInterface = $data{config}{interfaces}{aliasLookup}{$oAlias};
              foreach my $sourceIndex ((scalar @{$limitFrom} > 0 ? @{$limitFrom} : (($fromZone !~ /^(external)$/ && (scalar @{$data{config}{interfaces}{$fromZone}{$alias}{network}} > 0)) ? @{$data{config}{interfaces}{$fromZone}{$alias}{network}} : "")))
              {
                my $source = "";
                my %macInfo = ();
                if (length $sourceIndex > 0 && $sourceIndex >= 0)
                {
                  next if (!$data{config}{networks}{$sourceIndex}{active});

                  if ($data{config}{networks}{$sourceIndex}{type} eq "normal")
                  {
                    $source = "\${NETWORK_$sourceIndex}";
                  }
                  elsif ($data{config}{networks}{$sourceIndex}{type} eq "mac")
                  {
                    $macInfo{mac} = 1;
                    $macInfo{'mac-source'} = "\${NETWORK_$sourceIndex}";
                  }
                }
                foreach my $destIndex ((scalar @{$limitTo} > 0 ? @{$limitTo} : (($toZone !~ /^(external)$/ && (scalar @{$data{config}{interfaces}{$toZone}{$oAlias}{network}} > 0)) ? @{$data{config}{interfaces}{$toZone}{$oAlias}{network}} : "")))
                {
                  my $dest = "";
                  $dest = "\${NETWORK_$destIndex}" if (length $destIndex > 0 && $destIndex >= 0);
                  next if (!$data{config}{networks}{$destIndex}{active});

                  my %logInfo = ();
                  if ($log)
                  {
                    %logInfo = ("log" => 1, "log-prefix" => $logComment, "log-limit" => $normalLogLimit, "log-limit-rate" => $normalLogRate, "log-level" => $logLevel);
                  }
                  if ($reject)
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, %macInfo);

                        $filter->reject(chain => "forwarded" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - Rejected\n", %icmpInfo, "reject-with" => ($proto eq "tcp" ? "tcp-reset" : ""), %logInfo, %macInfo);
                      }
                    }
                  }
                  elsif ($queue)
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %macInfo);

                        $filter->queue(chain => "forwarded" . $chainEnding, i => $interface, o => $oInterface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo, %logInfo, %macInfo);

                        $nat->accept(chain => "post" . $chainEnding, o => $oInterface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - QUEUED\n", %icmpInfo);
                      }
                    }
                  }
                  elsif ($redirect)
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};
                      my $redirectPort = (length $pathNode->{redirectPort} > 0 ? $pathNode->{redirectPort} : $dport);

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->redirect(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - Redirected\n", "to-ports" => $redirectPort, %icmpInfo, %logInfo, %macInfo);

                        $filter->accept(chain => "incoming" . $chainEnding, i => $interface, d => "inInterfaceIP",
                                    s => $source, p => $proto, sport => "$sport", dport => "$redirectPort",
                                    preComment => "# Service $service, path $path - Redirected\n", %icmpInfo, %macInfo);
                      }
                    }
                  }
                  elsif ($dropNoLog)
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->drop(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %macInfo);

                        $filter->drop(chain => "forwarded" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - DropNoLog\n", %icmpInfo, %logInfo, %macInfo);
                      }
                    }
                  }
                  elsif ($dnat)
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};
                      my $dnatPort = (length $pathNode->{dnatPort} > 0 ? $pathNode->{dnatPort} : $dport);

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->dnat(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path - Dnated\n", "to-destination" => "$dnatDestination" . ($proto =~ /^(tcp|udp)$/ && length $dnatPort > 0 ? ":$dnatPort" : ""), %icmpInfo, %logInfo, %macInfo);

                        # now we have to potentially allow it through the firewall
                        if ($fromZone ne $dnatZone)
                        {
                          $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, d => $dnatDestination, o => $dnatInterface,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                    preComment => "# Service $service, path $path - Dnated\n", %icmpInfo, %macInfo);

                          # now we have to allow it out of the firewall (if the fromZone and dnatZone are the same then we have to snat.
                          $nat->accept(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                      s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                      preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                        }
                        else
                        {
                          $nat->snat(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                                      s => $source, p => $proto, sport => "$sport", dport => "$dnatPort",
                                      preComment => "# Service $service, path $path - Dnated into zone $dnatZone\n", "to-source" => "outInterfaceIP", %icmpInfo);
                        }
                      }
                    }
                  }
                  else  # handle as a normal ACCEPT.
                  {
                    foreach my $infoNode (@infos)
                    {
                      my $p = $infoNode->{proto};
                      my $dport = $infoNode->{dport};
                      my $sport = $infoNode->{sport};
                      my $icmpType = $infoNode->{icmpType};

                      foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                      {
                        my %icmpInfo = ();
                        if ($proto eq "icmp")
                        {
                          $icmpInfo{limit} = $icmpLimit;
                          $icmpInfo{'limit-rate'} = $icmpLimitRate;
                          $icmpInfo{'icmp-type'} = $icmpType;
                        }

                        my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                        $nat->accept(chain => "pre" . $chainEnding, i => $interface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path\n", %icmpInfo, %macInfo);

                        $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, o => $oInterface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path\n", %icmpInfo, %logInfo, %macInfo);

                        $nat->accept(chain => "post" . $chainEnding, o => $oInterface, d => $dest,
                                    s => $source, p => $proto, sport => "$sport", dport => "$dport",
                                    preComment => "# Service $service, path $path\n", %icmpInfo);
                      }
                    }
                  }
                } # end of destination
              } # end of source
            } # end of o interface
          } # end of i interface
        }
      }
      elsif ($path =~ /^(firewallToFirewall)$/)
      {
        # no limitTo and the specified interface is only used for it's IP Address.  The real interface is lo.
        foreach my $pathNode (@{$zonePaths{$path}})
        {
          next if (!$pathNode->{active});
          my $o = $pathNode->{o};
          my $redirect = $pathNode->{redirect};
          my $dnat = $pathNode->{dnat};
          my $dnatDestinationId = $pathNode->{dnatDestination};
          my $dnatDestination = "";
          $dnatDestination = "\${NETWORK_$dnatDestinationId}" if ($dnatDestinationId);
          my $dnatZone = $pathNode->{dnatZone};
          next if ($dnatZone =~ /^(internal|ipsec|dmz|dialin)$/ && !$data{config}{zones}{$dnatZone});
          my $dnatInterface = $data{config}{interfaces}{aliasLookup}{$pathNode->{dnatInterface}};
          my $log = $pathNode->{log};
          my $logComment = $pathNode->{logComment};
          my $logLevel = $pathNode->{logLevel};
          my @infos = $dataObj->returnPathServiceInfos($pathNode->{service});

          # find the toZone for the specified interface.
          foreach my $alias (@{$o})
          {
            my $oFound = 0;
            foreach my $zone (keys %{$data{config}{interfaces}})
            {
              next if ($zone eq "aliasLookup");
              if (exists $data{config}{interfaces}{$zone}{$alias})
              {
                $oFound = 1;
                $toZone = $zone;
              }
            }
            if (!$oFound)
            {
              die "Could not find interface '$alias' for path <$path>!\n";
            }

            next if (!$data{config}{interfaces}{$toZone}{$alias}{active});

            my $interface = $data{config}{interfaces}{aliasLookup}{$alias};
            my $dest = $data{config}{interfaces}{$toZone}{$alias}{ip};
            if ($dest eq "dynamic")
            {
              if ($data{config}{dynamicInterfacesMode} eq "lookupIP")
              {
                $dest = "\$" . uc($interface) . "IP";
              }
              else
              {
                $dest = "";
              }
            }

            my %logInfo = ();
            if ($log)
            {
              %logInfo = ("log" => 1, "log-prefix" => $logComment, "log-limit" => $normalLogLimit, "log-limit-rate" => $normalLogRate, "log-level" => $logLevel);
            }
            if ($redirect)
            {
              foreach my $infoNode (@infos)
              {
                my $p = $infoNode->{proto};
                my $dport = $infoNode->{dport};
                my $sport = $infoNode->{sport};
                my $icmpType = $infoNode->{icmpType};
                my $redirectPort = (length $pathNode->{redirectPort} > 0 ? $pathNode->{redirectPort} : $dport);

                foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                {
                  my %icmpInfo = ();
                  if ($proto eq "icmp")
                  {
                    $icmpInfo{limit} = $icmpLimit;
                    $icmpInfo{'limit-rate'} = $icmpLimitRate;
                    $icmpInfo{'icmp-type'} = $icmpType;
                  }

                  $nat->redirect(chain => "outgoingLOCAL", o => "lo", d => $dest,
                              s => $dest, p => $proto, sport => "$sport", dport => "$dport",
                              preComment => "# Service $service, path $path - Redirected\n", "to-ports" => $redirectPort, %icmpInfo, %logInfo);

                  $filter->accept(chain => "incomingLOCAL", i => "lo", d => $dest,
                              s => $dest, p => $proto, sport => "$sport", dport => "$redirectPort",
                              preComment => "# Service $service, path $path - Redirected\n", %icmpInfo);
                }
              }
            }
            elsif ($dnat)
            {
              foreach my $infoNode (@infos)
              {
                my $p = $infoNode->{proto};
                my $dport = $infoNode->{dport};
                my $sport = $infoNode->{sport};
                my $icmpType = $infoNode->{icmpType};
                my $dnatPort = (length $pathNode->{dnatPort} > 0 ? $pathNode->{dnatPort} : $dport);

                foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                {
                  my %icmpInfo = ();
                  if ($proto eq "icmp")
                  {
                    $icmpInfo{limit} = $icmpLimit;
                    $icmpInfo{'limit-rate'} = $icmpLimitRate;
                    $icmpInfo{'icmp-type'} = $icmpType;
                  }

                  my $chainEnding = ($proto eq "udp" ? "UDP" : ($proto eq "tcp" ? "TCP" : ($proto eq "icmp" ? "ICMP" : "OTHER")));

                  $nat->dnat(chain => "outgoingLOCAL", o => "lo", d => $dest,
                              p => $proto, sport => "$sport", dport => "$dport",
                              preComment => "# Service $service, path $path - Dnated\n", "to-destination" => "$dnatDestination" . ($proto =~ /^(tcp|udp)$/ && length $dnatPort > 0 ? ":$dnatPort" : ""), %icmpInfo, %logInfo);

                  # now we have to potentially allow it through the firewall
                  $filter->accept(chain => "forwarded" . $chainEnding, i => $interface, d => $dnatDestination, o => $dnatInterface,
                            s => "inInterfaceIP", p => $proto, sport => "$sport", dport => "$dnatPort",
                            preComment => "# Service $service, path $path - Dnated\n", %icmpInfo);
                  $nat->accept(chain => "post" . $chainEnding, o => $dnatInterface, d => $dnatDestination,
                              s => "outInterfaceIP", p => $proto, sport => "$sport", dport => "$dnatPort",
                              preComment => "# Service $service, path $path - Dnated into zone $dnatZone\n", %icmpInfo);
                }
              }
            }
            else  # handle as a normal ACCEPT.
            {
              foreach my $infoNode (@infos)
              {
                my $p = $infoNode->{proto};
                my $dport = $infoNode->{dport};
                my $sport = $infoNode->{sport};
                my $icmpType = $infoNode->{icmpType};

                foreach my $proto (($p eq "" ? ("tcp", "udp", "icmp", "") : $p))
                {
                  my %icmpInfo = ();
                  if ($proto eq "icmp")
                  {
                    $icmpInfo{limit} = $icmpLimit;
                    $icmpInfo{'limit-rate'} = $icmpLimitRate;
                    $icmpInfo{'icmp-type'} = $icmpType;
                  }

                  $nat->accept(chain => "outgoingLOCAL", o => "lo", d => $dest,
                              s => $dest, p => $proto, sport => "$sport", dport => "$dport",
                              preComment => "# Service $service, path $path\n", %icmpInfo);

                  $filter->accept(chain => "outgoingLOCAL", o => "lo", d => $dest,
                              s => $dest, p => $proto, sport => "$sport", dport => "$dport",
                              preComment => "# Service $service, path $path\n", %icmpInfo, %logInfo);

                  $nat->accept(chain => "postLOCAL", o => "lo", d => $dest,
                              s => $dest, p => $proto, sport => "$sport", dport => "$dport",
                              preComment => "# Service $service, path $path\n", %icmpInfo);
                }
              }
            }
          } # end of oAlias loop
        }
      }
      else
      {
        print "Path '$path' is not yet supported...\n";
      }
    }
  } # end of foreach service

  # do some Mangling of packets.
  if ($data{config}{mangleRules})
  {
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 21);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 23);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 25);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 53);
    $mangle->tos(chain => "OUTPUT", p => "udp", dport => 53);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 119);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 110);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 143);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 6667, preComment => "# IRC\n");
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 20, tosValue => 8);
    $mangle->tos(chain => "OUTPUT", p => "tcp", dport => 80, tosValue => 8);
  }

  # -------------------- Stop Defining your rules -----------------------------
}

# void stoppingRules(configObj, interfaceObj, filterObj, natObj, mangleObj, proc, networkObj)
#   Defines the rules/actions that need to be done when the firewall is shutting down.
sub stoppingRules
{
  my $self = shift;
  my %args = ( configObj => undef, interfaceObj => undef, filterObj => undef, natObj => undef, mangleObj => undef, proc => undef, networkObj => undef, @_ );
  my $config = $args{configObj};
  my $filter = $args{filterObj};
  my $nat = $args{natObj};
  my $mangle = $args{mangleObj};
  my $interfaces = $args{interfaceObj};
  my $proc = $args{proc};
  my $networkObj = $args{networkObj};

  # pull in the configuration values from the XML File.
  my $confObj = PCXFireWall::ConfigParser->new;
  my $dataObj = undef;
  my %data = ();
  eval { $dataObj = $confObj->parse(file => $self->{configFile}, module => "VeryTight2", moduleVersion => $self->{configFileVersion}); };
  if ($@)
  {
    die "Error parsing XML Config File '$self->{configFile}'!\n$@\n";
  }
  %data = %{$dataObj->{data}};

  # -------------------- Define your rules here -------------------------------

  # change log-prefix value
  $config->logPrefix($data{config}{logInfo}{prefix});  # This is the value used if you don't specify the log-prefix value.

  # set default post-comment to be a \n so we have some spacing.
  $config->postComment("\n");
  $config->preComment("");

  # Uncomment the following line if you want to use 'ip' instead of 'ifconfig'.
  $config->networkCommand($data{config}{networkCommand});

  # turn on ECN support by default
  $proc->sysNetIPv4(file => "tcp_ecn", value => "1");

  # Start responding to broadcast pings
  $proc->sysNetIPv4(file => "icmp_echo_ignore_broadcasts", value => "0", module => "Nat");

  # turn off synCookie and Martian logging support
  $proc->sysNetIPv4(file => "tcp_syncookies", value => "0");
  $proc->sysNetIPv4Conf(file => "log_martians", value => "0", iface => "all");

  # turn on/off ip forwarding, dynamic IP Addresses, icmp redirects, anti-spoofing blocks and source routing
  $proc->sysNetIPv4(file => "ip_forward", value => "0", module => "Nat", when => "Pre", required => "true");
  $proc->sysNetIPv4(file => "ip_dynaddr", value => "0", module => "Nat");
  $proc->sysNetIPv4Conf(file => "rp_filter", value => "0", module => "Nat", iface => "all");
  $proc->sysNetIPv4Conf(file => "accept_source_route", value => "1", module => "Nat", iface => "all");
  $proc->sysNetIPv4Conf(file => "accept_redirects", value => "1", module => "Nat", iface => "all");

  # turn off proxy_arp for those interfaces we enabled it on.
  foreach my $zone ("external", "internal", "ipsec", "dmz", "dialin")
  {
    # for each zone, check the interfaces defined in it, but only if the zone is active.
    if ($zone =~ /^(external)$/ || $data{config}{zones}{$zone})
    {
      foreach my $alias (keys %{$data{config}{interfaces}{$zone}})
      {
        next if (!$data{config}{interfaces}{$zone}{$alias}{active});

        my $interface = $data{config}{interfaces}{$zone}{$alias}{name};
        my $proxyarp = $data{config}{interfaces}{$zone}{$alias}{proxyarp};

        if ($proxyarp)
        {
          # disable proxy_arp support for this interface.
          $proc->sysNetIPv4Conf(file => "proxy_arp", value => "0", module => "Nat", iface => "$interface");
        }
      }
    }
  }

  # flush and delete all chains
  $filter->deleteChain(name => "All");
  $mangle->deleteChain(name => "All");
  $nat->deleteChain(name => "All");

  # set chain policies
  $filter->policy(chain => "FORWARD", policy => "DROP");  # If you set this to ACCEPT then your box is wide open!
  $filter->policy(chain => "INPUT", policy => "DROP");
  $filter->policy(chain => "OUTPUT", policy => "DROP");
  $mangle->policy(chain => "OUTPUT", policy => "ACCEPT");
  $mangle->policy(chain => "PREROUTING", policy => "ACCEPT");
  $mangle->policy(chain => "INPUT", policy => "ACCEPT");
  $mangle->policy(chain => "FORWARD", policy => "ACCEPT");
  $mangle->policy(chain => "POSTROUTING", policy => "ACCEPT");
  $nat->policy(chain => "PREROUTING", policy => "ACCEPT");
  $nat->policy(chain => "POSTROUTING", policy => "ACCEPT");
  $nat->policy(chain => "OUTPUT", policy => "ACCEPT");

  # -------------------- Stop Defining your rules -----------------------------
}

# void restartingRules(configObj)
#    Allows the user to specify the installPath to use if they want to
#    change it when we generate the restartfw script.
sub restartingRules
{
  my $self = shift;
  my %args = ( configObj => undef, @_ );
  my $config = $args{configObj};

  # -------------------- Define your rules here -------------------------------

  # if you want/need to change the installPath variable do it here.
  #$config->installPath("/var/etc/pcx_firewall");  # This is only used in the restartfw script.

  # -------------------- Stop Defining your rules -----------------------------
}

# ******************* User Configuration Ends here ****************************
# *****************************************************************************

sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;

  # define the configFile and configFileVersion values here.
  $self->{configFile} = "PCXFireWall/VeryTight2.xml";
  $self->{configFileVersion} = "2.1";

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://; # strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

1;
__END__

=head1 NAME

VeryTight2 - A PCXFireWall Template Rules Module.

=head1 SYNOPSIS

  This module is called and used by the generator script when you
  instruct it to build the startfw, stopfw and/or restartfw scripts
  using VeryTight as the Rules module with the following syntax:

    generator VeryTight2 [configFile.xml]

  where configFile.xml is optional.


=head1 DESCRIPTION

VeryTight2 is the Module that Defines a Template IPTables Rules
Module for the PCXFireWall Toolkit.  It is highly customizable.

=head1 FUNCTIONS

  scalar new(void)
    Creates a new instance of the PCXFireWall::VeryTight2
    object.

  void startingRules(configObj, interfaceObj, filterObj, natObj,
                     mangleObj, proc, modulesObj)
    Defines the rules/actions that need to be done when the firewall is
    starting up.

  void stoppingRules(configObj, interfaceObj, filterObj, natObj,
                     mangleObj, proc)
    Defines the rules/actions that need to be done when the firewall is
    shutting down.

  void restartingRules(configObj)
    Allows the user to specify the installPath to use if they want to
    change it when we generate the restartfw script.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 CONFIGURATION

  See the docs/VeryTight.html for complete documentation.

=head1 NOTES

  Do not enter any sites for allowedSites for DNS service as a fqdn
  (fully qualified domain name) as you will fail to resolve them when
  the rules are processed as your DNS rules will not have been put in
  place yet.

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1), PCXFireWall::Rules(3)

=cut
